print("import: 'sanic'")
import sanic

print("import: 'sanic.app'")
import sanic.app

print("import: 'sanic.blueprints'")
import sanic.blueprints

print("import: 'sanic.config'")
import sanic.config

print("import: 'sanic.constants'")
import sanic.constants

print("import: 'sanic.cookies'")
import sanic.cookies

print("import: 'sanic.exceptions'")
import sanic.exceptions

print("import: 'sanic.handlers'")
import sanic.handlers

print("import: 'sanic.log'")
import sanic.log

print("import: 'sanic.request'")
import sanic.request

print("import: 'sanic.response'")
import sanic.response

print("import: 'sanic.router'")
import sanic.router

print("import: 'sanic.server'")
import sanic.server

print("import: 'sanic.static'")
import sanic.static

print("import: 'sanic.testing'")
import sanic.testing

print("import: 'sanic.views'")
import sanic.views

print("import: 'sanic.websocket'")
import sanic.websocket

